@@ -62,7 +62,7 @@ angular.module('syncthing.core')
 
         $scope.folderDefaults = {
             selectedDevices: {},
-            type: "readwrite",
+            type: "sendreceive",
             rescanIntervalS: 3600,
             fsWatcherDelayS: 10,
             fsWatcherEnabled: true,
@@ -102,10 +102,10 @@ angular.module('syncthing.core')
             'touch': 'Update'
         };
         $scope.needIcons = {
-            'rm': 'trash-o',
-            'rmdir': 'trash-o',
-            'sync': 'arrow-circle-o-down',
-            'touch': 'asterisk'
+            'rm': 'far fa-fw fa-trash-alt',
+            'rmdir': 'far fa-fw fa-trash-alt',
+            'sync': 'far fa-fw arrow-alt-circle-down',
+            'touch': 'fas fa-fw fa-asterisk'
         };
 
         $scope.$on(Events.ONLINE, function () {
@@ -1097,8 +1097,14 @@ angular.module('syncthing.core')
             show: function() {
                 $scope.logging.refreshFacilities();
                 $scope.logging.timer = $timeout($scope.logging.fetch);
-                $('#logViewer').modal().on('hidden.bs.modal', function () {
+                var textArea = $('#logViewerText');
+                textArea.on("scroll", $scope.logging.onScroll);
+                $('#logViewer').modal().on('shown.bs.modal', function() {
+                    // Scroll to bottom.
+                    textArea.scrollTop(textArea[0].scrollHeight);
+                }).on('hidden.bs.modal', function () {
                     $timeout.cancel($scope.logging.timer);
+                    textArea.off("scroll", $scope.logging.onScroll);
                     $scope.logging.timer = null;
                     $scope.logging.entries = [];
                 });
@@ -1113,6 +1119,14 @@ angular.module('syncthing.core')
                     .success($scope.logging.refreshFacilities)
                     .error($scope.emitHTTPError);
             },
+            onScroll: function() {
+                var textArea = $('#logViewerText');
+                var scrollTop = textArea.prop('scrollTop');
+                var scrollHeight = textArea.prop('scrollHeight');
+                $scope.logging.paused = scrollHeight > (scrollTop + textArea.outerHeight());
+                // Browser events do not cause redraw, trigger manually.
+                $scope.$apply();
+            },
             timer: null,
             entries: [],
             paused: false,
@@ -1125,7 +1139,7 @@ angular.module('syncthing.core')
             },
             fetch: function() {
                 var textArea = $('#logViewerText');
-                if (textArea.is(":focus")) {
+                if ($scope.logging.paused) {
                     if (!$scope.logging.timer) return;
                     $scope.logging.timer = $timeout($scope.logging.fetch, 500);
                     return;
@@ -1139,11 +1153,14 @@ angular.module('syncthing.core')
                 $http.get(urlbase + '/system/log' + (last ? '?since=' + encodeURIComponent(last) : '')).success(function (data) {
                     if (!$scope.logging.timer) return;
                     $scope.logging.timer = $timeout($scope.logging.fetch, 2000);
-                    if (!textArea.is(":focus")) {
+                    if (!$scope.logging.paused) {
                         if (data.messages) {
                             $scope.logging.entries.push.apply($scope.logging.entries, data.messages);
+                            // Wait for the text area to be redrawn, adding new lines, and then scroll to bottom.
+                            $timeout(function() {
+                                textArea.scrollTop(textArea[0].scrollHeight);
+                            });
                         }
-                        textArea.scrollTop(textArea[0].scrollHeight);
                     }
                 });
             }
@@ -1466,6 +1483,16 @@ angular.module('syncthing.core')
             $http.post(urlbase + '/system/error/clear');
         };
 
+        $scope.fsWatcherErrorMap = function () {
+            var errs = {}
+            $.each($scope.folders, function (id, cfg) {
+                if (cfg.fsWatcherEnabled && $scope.model[cfg.id] && $scope.model[id].watchError && !cfg.paused && $scope.folderStatus(cfg) !== 'stopped') {
+                    errs[id] = $scope.model[id].watchError;
+                }
+            });
+            return errs;
+        }
+
         $scope.friendlyDevices = function (str) {
             for (var i = 0; i < $scope.devices.length; i++) {
                 var cfg = $scope.devices[i];
@@ -1506,13 +1533,13 @@ angular.module('syncthing.core')
             $scope.currentFolder.path = pathJoin($scope.config.options.defaultFolderPath, newvalue);
         });
 
-        $scope.$watch('currentFolder.fsWatcherEnabled', function (newvalue) {
-            if (newvalue) {
+        $scope.fsWatcherToggled = function () {
+            if ($scope.currentFolder.fsWatcherEnabled) {
                 $scope.currentFolder.rescanIntervalS = 3600;
             } else {
                 $scope.currentFolder.rescanIntervalS = 60;
             }
-        });
+        };
 
         $scope.loadFormIntoScope = function (form) {
             console.log('loadFormIntoScope',form.$name);
@@ -1548,7 +1575,7 @@ angular.module('syncthing.core')
         $scope.editFolder = function (folderCfg) {
             $scope.editingExisting = true;
             $scope.currentFolder = angular.copy(folderCfg);
-            if ($scope.currentFolder.path.slice(-1) === $scope.system.pathSeparator) {
+            if ($scope.currentFolder.path.length > 1 && $scope.currentFolder.path.slice(-1) === $scope.system.pathSeparator) {
                 $scope.currentFolder.path = $scope.currentFolder.path.slice(0, -1);
             }
             $scope.currentFolder.selectedDevices = {};
@@ -1760,6 +1787,9 @@ angular.module('syncthing.core')
         };
 
         $scope.folderLabel = function (folderID) {
+            if (!$scope.folders[folderID]) {
+                return folderID;
+            }
             var label = $scope.folders[folderID].label;
             return label.length > 0 ? label : folderID;
         }
@@ -2245,6 +2275,9 @@ angular.module('syncthing.core')
         };
 
         $scope.sizeOf = function (dict) {
+            if (dict === undefined) {
+                return 0;
+            }
             return Object.keys(dict).length;
         };
 
