@@ -32,7 +32,7 @@ import (
 
 const (
 	OldestHandledVersion = 10
-	CurrentVersion       = 27
+	CurrentVersion       = 28
 	MaxRescanIntervalS   = 365 * 24 * 60 * 60
 )
 
@@ -315,6 +315,9 @@ func (cfg *Configuration) clean() error {
 	if cfg.Version == 26 {
 		convertV26V27(cfg)
 	}
+	if cfg.Version == 27 {
+		convertV27V28(cfg)
+	}
 
 	// Build a list of available devices
 	existingDevices := make(map[protocol.DeviceID]bool)
@@ -322,13 +325,18 @@ func (cfg *Configuration) clean() error {
 		existingDevices[device.DeviceID] = true
 	}
 
-	// Ensure that the device list is free from duplicates
+	// Ensure that the device list is
+	// - free from duplicates
+	// - sorted by ID
 	cfg.Devices = ensureNoDuplicateDevices(cfg.Devices)
-
 	sort.Sort(DeviceConfigurationList(cfg.Devices))
-	// Ensure that any loose devices are not present in the wrong places
-	// Ensure that there are no duplicate devices
-	// Ensure that the versioning configuration parameter map is not nil
+
+	// Ensure that the folder list is sorted by ID
+	sort.Sort(FolderConfigurationList(cfg.Folders))
+	// Ensure that in all folder configs
+	// - any loose devices are not present in the wrong places
+	// - there are no duplicate devices
+	// - the versioning configuration parameter map is not nil
 	for i := range cfg.Folders {
 		cfg.Folders[i].Devices = ensureExistingDevices(cfg.Folders[i].Devices, existingDevices)
 		cfg.Folders[i].Devices = ensureNoDuplicateFolderDevices(cfg.Folders[i].Devices)
@@ -374,6 +382,21 @@ func (cfg *Configuration) clean() error {
 	return nil
 }
 
+// DeviceMap returns a map of device ID to device configuration for the given configuration.
+func (cfg *Configuration) DeviceMap() map[protocol.DeviceID]DeviceConfiguration {
+	m := make(map[protocol.DeviceID]DeviceConfiguration, len(cfg.Devices))
+	for _, dev := range cfg.Devices {
+		m[dev.DeviceID] = dev
+	}
+	return m
+}
+
+func convertV27V28(cfg *Configuration) {
+	// Show a notification about enabling filesystem watching
+	cfg.Options.UnackedNotificationIDs = append(cfg.Options.UnackedNotificationIDs, "fsWatcherNotification")
+	cfg.Version = 28
+}
+
 func convertV26V27(cfg *Configuration) {
 	for i := range cfg.Folders {
 		f := &cfg.Folders[i]
