@@ -25,7 +25,8 @@ var (
 // The BasicFilesystem implements all aspects by delegating to package os.
 // All paths are relative to the root and cannot (should not) escape the root directory.
 type BasicFilesystem struct {
-	root string
+	root                 string
+	rootSymlinkEvaluated string
 }
 
 func newBasicFilesystem(root string) *BasicFilesystem {
@@ -52,21 +53,34 @@ func newBasicFilesystem(root string) *BasicFilesystem {
 		}
 	}
 
+	rootSymlinkEvaluated, err := filepath.EvalSymlinks(root)
+	if err != nil {
+		rootSymlinkEvaluated = root
+	}
+
+	return &BasicFilesystem{
+		root:                 adjustRoot(root),
+		rootSymlinkEvaluated: adjustRoot(rootSymlinkEvaluated),
+	}
+}
+
+func adjustRoot(root string) string {
 	// Attempt to enable long filename support on Windows. We may still not
 	// have an absolute path here if the previous steps failed.
 	if runtime.GOOS == "windows" {
 		if filepath.IsAbs(root) && !strings.HasPrefix(root, `\\`) {
 			root = `\\?\` + root
 		}
-		// If we're not on Windows, we want the path to end with a slash to
-		// penetrate symlinks. On Windows, paths must not end with a slash.
-	} else if root[len(root)-1] != filepath.Separator {
+		return root
+	}
+
+	// If we're not on Windows, we want the path to end with a slash to
+	// penetrate symlinks. On Windows, paths must not end with a slash.
+	if root[len(root)-1] != filepath.Separator {
 		root = root + string(filepath.Separator)
 	}
 
-	return &BasicFilesystem{
-		root: root,
-	}
+	return root
 }
 
 // rooted expands the relative path to the full path that is then used with os
@@ -74,8 +88,20 @@ func newBasicFilesystem(root string) *BasicFilesystem {
 // directory, this returns an error, to prevent accessing files that are not in the
 // shared directory.
 func (f *BasicFilesystem) rooted(rel string) (string, error) {
+	return rooted(rel, f.root)
+}
+
+// rootedSymlinkEvaluated does the same as rooted, but the returned path will not
+// contain any symlinks.  package. If the relative path somehow causes the final
+// path to escape the root directory, this returns an error, to prevent accessing
+// files that are not in the shared directory.
+func (f *BasicFilesystem) rootedSymlinkEvaluated(rel string) (string, error) {
+	return rooted(rel, f.rootSymlinkEvaluated)
+}
+
+func rooted(rel, root string) (string, error) {
 	// The root must not be empty.
-	if f.root == "" {
+	if root == "" {
 		return "", ErrInvalidFilename
 	}
 
@@ -83,7 +109,7 @@ func (f *BasicFilesystem) rooted(rel string) (string, error) {
 
 	// The expected prefix for the resulting path is the root, with a path
 	// separator at the end.
-	expectedPrefix := filepath.FromSlash(f.root)
+	expectedPrefix := filepath.FromSlash(root)
 	if !strings.HasSuffix(expectedPrefix, pathSep) {
 		expectedPrefix += pathSep
 	}
@@ -97,7 +123,7 @@ func (f *BasicFilesystem) rooted(rel string) (string, error) {
 	// The supposedly correct path is the one filepath.Join will return, as
 	// it does cleaning and so on. Check that one first to make sure no
 	// obvious escape attempts have been made.
-	joined := filepath.Join(f.root, rel)
+	joined := filepath.Join(root, rel)
 	if rel == "." && !strings.HasSuffix(joined, pathSep) {
 		joined += pathSep
 	}
@@ -109,7 +135,15 @@ func (f *BasicFilesystem) rooted(rel string) (string, error) {
 }
 
 func (f *BasicFilesystem) unrooted(path string) string {
-	return strings.TrimPrefix(strings.TrimPrefix(path, f.root), string(PathSeparator))
+	return rel(path, f.root)
+}
+
+func (f *BasicFilesystem) unrootedSymlinkEvaluated(path string) string {
+	return rel(path, f.rootSymlinkEvaluated)
+}
+
+func rel(path, prefix string) string {
+	return strings.TrimPrefix(strings.TrimPrefix(path, prefix), string(PathSeparator))
 }
 
 func (f *BasicFilesystem) Chmod(name string, mode FileMode) error {
